@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -273,6 +273,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
     switch( handle_type )
     {
       case SQL_HANDLE_ENV:
+      case SQL_HANDLE_SENV:
         {
             DMHENV environment;
             char pooling_string[ 128 ];
@@ -320,6 +321,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
             environment -> state = STATE_E1;
             environment -> requested_version = requested_version;
+            environment -> version_set = !!requested_version;
         	environment -> sql_driver_count = -1;
 
             /*
@@ -390,7 +392,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                         ERROR_HY009, NULL, 
                         SQL_OV_ODBC3 );
 
-                return function_return( SQL_HANDLE_ENV, environment, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_ENV, environment, SQL_ERROR );
             }
 
             /*
@@ -411,7 +413,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
                 *output_handle = SQL_NULL_HDBC;
 
-                return function_return( SQL_HANDLE_ENV, environment, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_ENV, environment, SQL_ERROR );
             }
 
             connection = __alloc_dbc();
@@ -429,7 +431,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
                 *output_handle = SQL_NULL_HDBC;
 
-                return function_return( SQL_HANDLE_ENV, environment, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_ENV, environment, SQL_ERROR );
             }
 
             /*
@@ -447,9 +449,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
             connection -> cursors = SQL_CUR_DEFAULT;
             connection -> login_timeout = SQL_LOGIN_TIMEOUT_DEFAULT;
             connection -> login_timeout_set = 0;
-            connection -> auto_commit = 0;
-            connection -> auto_commit_set = 0;
-            connection -> auto_commit = 0;
+            connection -> auto_commit = SQL_AUTOCOMMIT_ON;
             connection -> auto_commit_set = 0;
             connection -> async_enable = 0;
             connection -> async_enable_set = 0;
@@ -555,7 +555,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                         ERROR_HY009, NULL, 
                         connection -> environment -> requested_version  );
 
-                return function_return( SQL_HANDLE_DBC, connection , SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_DBC, connection , SQL_ERROR );
             }
 
             if ( connection -> state == STATE_C1 ||
@@ -574,7 +574,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
                 *output_handle = SQL_NULL_HSTMT;
 
-                return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
             }
 
             statement = __alloc_stmt();
@@ -592,14 +592,14 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
                 *output_handle = SQL_NULL_HSTMT;
 
-                return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
             }
 
             /*
              * pass the call on
              */
 
-            if ( requested_version == SQL_OV_ODBC3 )
+            if ( requested_version >= SQL_OV_ODBC3 )
             {
                 if ( CHECK_SQLALLOCHANDLE( connection ))
                 {
@@ -638,7 +638,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
                     *output_handle = SQL_NULL_HSTMT;
 
-                    return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
                 }
             }
             else
@@ -680,7 +680,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
                     *output_handle = SQL_NULL_HSTMT;
                 
-                    return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
                 }
             }
 
@@ -694,9 +694,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                 if ( connection -> state == STATE_C4 )
                     connection -> state = STATE_C5;
 
-                connection -> statement_count ++;
-
-                statement -> connection = connection;
+                __register_stmt ( connection, statement );
 
                 *output_handle = (SQLHANDLE) statement;
 
@@ -745,7 +743,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
                             __release_stmt( statement );
 
-                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR, DEFER_R1 );
                         }
                         statement -> implicit_ard = statement -> ard;
                         statement -> ard -> implicit = 1;
@@ -789,7 +787,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
                             *output_handle = SQL_NULL_HSTMT;
 
-                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR, DEFER_R1 );
                         }
                         statement -> implicit_apd = statement -> apd;
                         statement -> apd -> implicit = 1;
@@ -833,7 +831,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
                             *output_handle = SQL_NULL_HSTMT;
 
-                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR, DEFER_R1 );
                         }
                         statement -> implicit_ird = statement -> ird;
                         statement -> ird -> implicit = 1;
@@ -873,9 +871,11 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                                     ERROR_HY013, NULL,
                                     connection -> environment -> requested_version );
 
+                            __release_stmt( statement );
+
                             *output_handle = SQL_NULL_HSTMT;
 
-                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR, DEFER_R1 );
                         }
                         statement -> implicit_ipd = statement -> ipd;
                         statement -> ipd -> implicit = 1;
@@ -922,7 +922,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
                             __release_stmt( statement );
 
-                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR, DEFER_R1 );
                         }
                         statement -> implicit_ard = statement -> ard;
                         statement -> ard -> implicit = 1;
@@ -966,7 +966,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
                             *output_handle = SQL_NULL_HSTMT;
 
-                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR, DEFER_R1 );
                         }
                         statement -> implicit_apd = statement -> apd;
                         statement -> apd -> implicit = 1;
@@ -1010,7 +1010,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
                             *output_handle = SQL_NULL_HSTMT;
 
-                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR, DEFER_R1 );
                         }
                         statement -> implicit_ird = statement -> ird;
                         statement -> ird -> implicit = 1;
@@ -1050,9 +1050,11 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                                     ERROR_HY013, NULL,
                                     connection -> environment -> requested_version );
 
+                            __release_stmt( statement );
+
                             *output_handle = SQL_NULL_HSTMT;
 
-                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR, DEFER_R1 );
                         }
                         statement -> implicit_ipd = statement -> ipd;
                         statement -> ipd -> implicit = 1;
@@ -1090,7 +1092,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                                UODBC_STATS_TYPE_HSTMT, (void *)1);
 #endif
 
-            return function_return( SQL_HANDLE_DBC, connection, ret );
+            return function_return( SQL_HANDLE_DBC, connection, ret, DEFER_R1 );
         }
         break;
 
@@ -1144,7 +1146,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                     ERROR_HY009, NULL, 
                     connection -> environment -> requested_version  );
 
-            return function_return( SQL_HANDLE_DBC, connection , SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection , SQL_ERROR );
         }
 
         if ( connection -> state == STATE_C1 ||
@@ -1163,7 +1165,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
             *output_handle = SQL_NULL_HDESC;
 
-            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
 
         descriptor = __alloc_desc();
@@ -1181,7 +1183,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
             *output_handle = SQL_NULL_HDESC;
 
-            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
 
         /*
@@ -1215,7 +1217,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 
             *output_handle = SQL_NULL_HDESC;
 
-            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
 
         if ( SQL_SUCCEEDED( ret ))
@@ -1252,7 +1254,7 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
                            (void *)1);
 #endif
 
-        return function_return( SQL_HANDLE_DBC, connection, ret );
+        return function_return( SQL_HANDLE_DBC, connection, ret, DEFER_R1 );
         }
         break;
 
@@ -1260,24 +1262,30 @@ SQLRETURN __SQLAllocHandle( SQLSMALLINT handle_type,
 		if ( __validate_env( (DMHENV) input_handle ))
 		{
 			DMHENV environment = (DMHENV) input_handle;
+
+            thread_protect( SQL_HANDLE_ENV, environment );
+
 			__post_internal_error( &environment -> error,
 						ERROR_HY092, NULL,
 						environment -> requested_version );
 
-			return function_return( SQL_HANDLE_ENV, environment, SQL_ERROR );
+			return function_return_nodrv( SQL_HANDLE_ENV, environment, SQL_ERROR );
 		}
 		else if ( __validate_dbc( (DMHDBC) input_handle ))
 		{
 			DMHDBC connection = (DMHDBC) input_handle;
+
+            thread_protect( SQL_HANDLE_DBC, connection );
+
 			__post_internal_error( &connection -> error,
 					ERROR_HY092, NULL,
 					connection -> environment -> requested_version );
 	
-			return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+			return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
 		}
 		else
 		{
-			return SQL_ERROR;
+			return SQL_INVALID_HANDLE;
 		}
 		break;
     }
