@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -188,6 +188,7 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
     switch( handle_type )
     {
       case SQL_HANDLE_ENV:
+      case SQL_HANDLE_SENV:
         {
             DMHENV environment = (DMHENV)handle;
 
@@ -239,13 +240,16 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
                         ERROR_HY010, NULL,
                         environment -> requested_version );
 
-            	thread_release( SQL_HANDLE_ENV, environment );
-
-                return function_return( SQL_HANDLE_ENV, environment, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_ENV, environment, SQL_ERROR );
             }
 
             thread_release( SQL_HANDLE_ENV, environment );
 
+            /*
+             * release any pooled connections that are using this environment
+             */
+            __strip_from_pool( environment );
+
             __release_env( environment );
             return SQL_SUCCESS;
         }
@@ -306,9 +310,7 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
                         ERROR_HY010, NULL,
                         connection -> environment -> requested_version );
 
-            	thread_release( SQL_HANDLE_ENV, environment );
-
-                return function_return( SQL_HANDLE_ENV, environment, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_ENV, environment, SQL_ERROR );
             }
 
             environment -> connection_count --;
@@ -397,7 +399,10 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
                     statement -> state == STATE_S9 ||
                     statement -> state == STATE_S10 ||
                     statement -> state == STATE_S11 ||
-                    statement -> state == STATE_S12 )
+                    statement -> state == STATE_S12 ||
+                    statement -> state == STATE_S13 ||
+                    statement -> state == STATE_S14 ||
+                    statement -> state == STATE_S15 )
             {
                 dm_log_write( __FILE__,
                         __LINE__,
@@ -409,9 +414,7 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
                           ERROR_HY010, NULL,
                           statement -> connection -> environment -> requested_version );
 
-                thread_release( SQL_HANDLE_STMT, statement );
-				
-                return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
             }
 
             if ( !CHECK_SQLFREEHANDLE( statement -> connection ))
@@ -428,7 +431,7 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
                             ERROR_IM001, NULL,
                             statement -> connection -> environment -> requested_version );
 
-                    return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+                    return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
                 }
                 else
                 {
@@ -446,16 +449,6 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
 
             if ( SQL_SUCCEEDED( ret ))
             {
-                /*
-                 * break any association
-                 */
-
-                if ( statement -> ard ) {
-                    statement -> ard -> associated_with = NULL;
-                }
-                if ( statement -> apd ) {
-                    statement -> apd -> associated_with = NULL;
-                }
                 /*
                  * release the implicit descriptors, 
 				 * this matches the tests in SQLAllocHandle
@@ -500,7 +493,7 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
                         connection -> msg );
             }
 
-            return function_return( IGNORE_THREAD, connection, ret ); 
+            return function_return( IGNORE_THREAD, connection, ret, DEFER_R0 );
         }
         break;
 
@@ -548,7 +541,7 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
 						ERROR_HY017, NULL,
 						connection -> environment -> requested_version );
 		
-				return function_return( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
+				return function_return_nodrv( IGNORE_THREAD, descriptor, SQL_ERROR );
 			}
 		
             thread_protect( SQL_HANDLE_DESC, descriptor );
@@ -565,9 +558,7 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
                         ERROR_IM001, NULL,
                         connection -> environment -> requested_version );
 
-            	thread_release( SQL_HANDLE_DESC, descriptor );
-
-                return function_return( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
             }
             else
             {
@@ -596,7 +587,7 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
                         ERROR_HY010, NULL,
                         descriptor -> connection -> environment -> requested_version );
 
-                return function_return( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
+                return function_return( SQL_HANDLE_DESC, descriptor, SQL_ERROR, DEFER_R0 );
             }
 
             thread_release( SQL_HANDLE_DESC, descriptor );
@@ -619,7 +610,7 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
                                UODBC_STATS_TYPE_HDESC, (void *)-1);
 #endif
 
-            return function_return( IGNORE_THREAD, connection, SQL_SUCCESS ); 
+            return function_return( IGNORE_THREAD, connection, SQL_SUCCESS, DEFER_R0 );
         }
         break;
 
@@ -627,7 +618,7 @@ SQLRETURN __SQLFreeHandle( SQLSMALLINT handle_type,
         /*
          * there is nothing to report a error on
          */
-        return SQL_ERROR;
+        return SQL_INVALID_HANDLE;
     }
 }
 
