@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -204,7 +204,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
             }
             else
             {
-                *((SQLINTEGER*)value) = SQL_OPT_TRACE_ON;
+                *((SQLINTEGER*)value) = SQL_OPT_TRACE_OFF;
             }
         }
 
@@ -309,7 +309,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
                 ERROR_HY010, NULL,
                 connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
     }
 
     if ( connection -> state == STATE_C2 )
@@ -325,7 +325,89 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
           case SQL_ATTR_ASYNC_ENABLE:
             break;
 
+          case SQL_ATTR_PACKET_SIZE:
+            if ( connection -> packet_size_set )
+                break;
+          case SQL_ATTR_QUIET_MODE:
+            if ( connection -> quite_mode_set )
+                break;
+
           default:
+            {
+                struct save_attr *sa = connection -> save_attr;
+                while (sa)
+                {
+                    if (sa -> attr_type == attribute)
+                    {
+                        SQLRETURN rc = SQL_SUCCESS;
+                        if (sa -> str_len == SQL_NTS || sa -> str_len > 0)
+                        {
+                            SQLLEN realLen = sa->str_attr ? strlen(sa->str_attr) : 0;
+                            if(value && sa->str_attr)
+                            {
+                                strncpy(value, sa->str_attr, buffer_length - 1);
+                                ((SQLCHAR*)value)[buffer_length - 1] = 0;
+                            }
+                            if(string_length)
+                            {
+                                *string_length = realLen;
+                            }
+                            if(realLen > buffer_length - 1)
+                            {
+                                __post_internal_error( &connection -> error,
+                                ERROR_01004, NULL,
+                                connection -> environment -> requested_version );
+                                rc = SQL_SUCCESS_WITH_INFO;
+                            }
+                        }
+                        else if(buffer_length >= sizeof(SQLLEN))
+                        {
+                            *(SQLLEN*)value = sa -> int_attr;
+                            if(string_length)
+                            {
+                                *string_length = sizeof(SQLLEN);
+                            }
+                        }
+                        else if(sa -> str_len >= SQL_IS_SMALLINT && sa -> str_len <= SQL_IS_POINTER)
+                        {
+                            SQLLEN length = 0;
+                            switch (sa -> str_len)
+                            {
+                            case SQL_IS_SMALLINT:
+                                *(SQLSMALLINT*)value = sa->int_attr;
+                                length = sizeof(SQLSMALLINT);
+                                break;
+                            case SQL_IS_USMALLINT:
+                                *(SQLUSMALLINT*)value = sa->int_attr;
+                                length = sizeof(SQLUSMALLINT);
+                                break;
+                            case SQL_IS_INTEGER:
+                                *(SQLINTEGER*)value = sa->int_attr;
+                                length = sizeof(SQLINTEGER);
+                                break;
+                            case SQL_IS_UINTEGER:
+                                *(SQLUINTEGER*)value = sa->int_attr;
+                                length = sizeof(SQLUINTEGER);
+                                break;
+                            case SQL_IS_POINTER:
+                                *(SQLPOINTER**)value = sa->int_attr;
+                                length = sizeof(SQLPOINTER);
+                                break;
+                            }
+                            if(string_length)
+                            {
+                                *string_length = length;
+                            }
+                        }
+                        else
+                        {
+                            memcpy(value, &sa->int_attr, buffer_length);
+                        }
+                        return function_return_nodrv( SQL_HANDLE_DBC, connection, rc );
+                    }
+                    sa = sa -> next;
+                }
+            }
             dm_log_write( __FILE__, 
                     __LINE__, 
                     LOG_INFO, 
@@ -336,7 +418,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
                     ERROR_08003, NULL,
                     connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
     }
 
@@ -388,7 +470,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
         break;
 
       case SQL_ATTR_ODBC_CURSORS:
-        *((SQLINTEGER*)value) = connection -> cursors;
+        *((SQLULEN*)value) = connection -> cursors;
         type = 1;
         break;
 
@@ -412,7 +494,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
         }
         else
         {
-            *((SQLINTEGER*)value) = connection -> async_enable;
+            *((SQLULEN*)value) = connection -> async_enable;
             type = 1;
         }
         break;
@@ -556,7 +638,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
                 LOG_INFO, 
                 connection -> msg );
 
-        return function_return( SQL_HANDLE_DBC, connection, ret );
+        return function_return_nodrv( SQL_HANDLE_DBC, connection, ret );
     }
     else
     {
@@ -590,7 +672,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
                                 ERROR_HY092, NULL,
                                 connection -> environment -> requested_version );
 
-                        return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
                     }
 
                     switch( attribute )
@@ -617,7 +699,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
                       case SQL_ATTR_TRANSLATE_LIB:
                         if ( SQL_SUCCEEDED( ret ) && value && s1 )
                         {
-                            unicode_to_ansi_copy( value, buffer_length, s1, SQL_NTS, connection );
+                            unicode_to_ansi_copy( value, buffer_length, s1, SQL_NTS, connection, NULL );
                         }
 						if ( SQL_SUCCEEDED( ret ) && string_length ) 
 						{
@@ -643,7 +725,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
                             ERROR_IM001, NULL,
                             connection -> environment -> requested_version );
 
-                    return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
                 }
             }
             else
@@ -666,7 +748,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
                     connection -> driver_dbc,
                     attribute,
                     s1 ? s1 : value,
-                    buffer_length,
+                    s1 ? sizeof( SQLWCHAR ) * buffer_length : buffer_length,
                     string_length );
 
                 switch( attribute )
@@ -676,7 +758,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
                   case SQL_ATTR_TRANSLATE_LIB:
                     if ( SQL_SUCCEEDED( ret ) && value && s1 )
                     {
-                        unicode_to_ansi_copy( value, buffer_length, s1, SQL_NTS, connection );
+                        unicode_to_ansi_copy( value, buffer_length, s1, SQL_NTS, connection, NULL );
                     }
 					if ( SQL_SUCCEEDED( ret ) && string_length ) 
 					{
@@ -714,7 +796,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
                                 ERROR_HY092, NULL,
                                 connection -> environment -> requested_version );
 
-                        return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
                     }
 
                     ret = SQLGETCONNECTOPTION( connection,
@@ -734,7 +816,7 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
                             ERROR_IM001, NULL,
                             connection -> environment -> requested_version );
 
-                    return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
                 }
             }
             else
@@ -762,6 +844,6 @@ SQLRETURN SQLGetConnectAttr( SQLHDBC connection_handle,
         }
 
 
-        return function_return( SQL_HANDLE_DBC, connection, ret );
+        return function_return( SQL_HANDLE_DBC, connection, ret, DEFER_R3 );
     }
 }
