@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -215,40 +215,6 @@ SQLRETURN SQLSpecialColumns( SQLHSTMT statement_handle,
 
     thread_protect( SQL_HANDLE_STMT, statement );
 
-    /*
-     * Check the SQL_ATTR_METADATA_ID settings
-     */
-
-    if ( statement -> metadata_id == SQL_TRUE &&
-            schema_name == NULL )
-    {
-        __post_internal_error( &statement -> error,
-                ERROR_HY009, NULL,
-                statement -> connection -> environment -> requested_version );
-
-        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
-    }
-
-    if ( table_name == NULL )
-    {
-        __post_internal_error( &statement -> error,
-                ERROR_HY009, NULL,
-                statement -> connection -> environment -> requested_version );
-
-        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
-    }
-
-    if (( name_length1 < 0 && name_length1 != SQL_NTS ) ||
-            ( name_length2 < 0 && name_length3 != SQL_NTS ) ||
-            ( name_length3 < 0 && name_length3 != SQL_NTS ))
-    {
-        __post_internal_error( &statement -> error,
-                ERROR_HY090, NULL,
-                statement -> connection -> environment -> requested_version );
-
-        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
-    }
-
     if ( identifier_type != SQL_BEST_ROWID &&
             identifier_type != SQL_ROWVER )
     {
@@ -262,7 +228,49 @@ SQLRETURN SQLSpecialColumns( SQLHSTMT statement_handle,
                 ERROR_HY097, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
+    }
+
+    if ( name_length1 < 0 && name_length1 != SQL_NTS ||
+         name_length2 < 0 && name_length2 != SQL_NTS )
+    {
+        __post_internal_error( &statement -> error,
+                ERROR_HY090, NULL,
+                statement -> connection -> environment -> requested_version );
+
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
+    }
+
+    if ( table_name == NULL )
+    {
+        __post_internal_error( &statement -> error,
+                ERROR_HY009, NULL,
+                statement -> connection -> environment -> requested_version );
+
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
+    }
+
+    if ( name_length3 < 0 && name_length3 != SQL_NTS )
+    {
+        __post_internal_error( &statement -> error,
+                ERROR_HY090, NULL,
+                statement -> connection -> environment -> requested_version );
+
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
+    }
+
+    /*
+     * Check the SQL_ATTR_METADATA_ID settings
+     */
+
+    if ( statement -> metadata_id == SQL_TRUE &&
+            schema_name == NULL )
+    {
+        __post_internal_error( &statement -> error,
+                ERROR_HY009, NULL,
+                statement -> connection -> environment -> requested_version );
+
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 
     if ( scope != SQL_SCOPE_CURROW &&
@@ -279,7 +287,7 @@ SQLRETURN SQLSpecialColumns( SQLHSTMT statement_handle,
                 ERROR_HY098, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR ); 
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR ); 
     }
 
     if ( nullable != SQL_NO_NULLS &&
@@ -295,7 +303,7 @@ SQLRETURN SQLSpecialColumns( SQLHSTMT statement_handle,
                 ERROR_HY099, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 
     /*
@@ -321,11 +329,14 @@ SQLRETURN SQLSpecialColumns( SQLHSTMT statement_handle,
                 ERROR_24000, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR ); 
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR ); 
     }
     else if ( statement -> state == STATE_S8 ||
             statement -> state == STATE_S9 ||
-            statement -> state == STATE_S10 )
+            statement -> state == STATE_S10 ||
+            statement -> state == STATE_S13 ||
+            statement -> state == STATE_S14 ||
+            statement -> state == STATE_S15 )
     {
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -337,7 +348,7 @@ SQLRETURN SQLSpecialColumns( SQLHSTMT statement_handle,
                 ERROR_HY010, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 
     if ( statement -> state == STATE_S11 ||
@@ -355,13 +366,14 @@ SQLRETURN SQLSpecialColumns( SQLHSTMT statement_handle,
                     ERROR_HY010, NULL,
                     statement -> connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
     }
 
     if ( statement -> connection -> unicode_driver )
     {
         SQLWCHAR *s1, *s2, *s3;
+        int wlen;
 
         if ( !CHECK_SQLSPECIALCOLUMNSW( statement -> connection ))
         {
@@ -375,12 +387,15 @@ SQLRETURN SQLSpecialColumns( SQLHSTMT statement_handle,
                     ERROR_IM001, NULL,
                     statement -> connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
 
-        s1 = ansi_to_unicode_alloc( catalog_name, name_length1, statement -> connection );
-        s2 = ansi_to_unicode_alloc( schema_name, name_length2, statement -> connection );
-        s3 = ansi_to_unicode_alloc( table_name, name_length3, statement -> connection );
+        s1 = ansi_to_unicode_alloc( catalog_name, name_length1, statement -> connection, &wlen );
+        name_length1 = wlen;
+        s2 = ansi_to_unicode_alloc( schema_name, name_length2, statement -> connection, &wlen );
+        name_length2 = wlen;
+        s3 = ansi_to_unicode_alloc( table_name, name_length3, statement -> connection, &wlen );
+        name_length3 = wlen;
 
         ret = SQLSPECIALCOLUMNSW( statement -> connection ,
                 statement -> driver_stmt,
@@ -415,7 +430,7 @@ SQLRETURN SQLSpecialColumns( SQLHSTMT statement_handle,
                     ERROR_IM001, NULL,
                     statement -> connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
 
         ret = SQLSPECIALCOLUMNS( statement -> connection ,
@@ -474,5 +489,5 @@ SQLRETURN SQLSpecialColumns( SQLHSTMT statement_handle,
                 statement -> msg );
     }
 
-    return function_return( SQL_HANDLE_STMT, statement, ret );
+    return function_return( SQL_HANDLE_STMT, statement, ret, DEFER_R1 );
 }
