@@ -20,6 +20,7 @@
  * -----------------------------------------------
  * Peter Harvey		- pharvey@codebydesign.com
  **************************************************/
+#include <config.h>
 #include <odbcinstext.h>
 
 BOOL SQLInstallDriverEx(		LPCSTR	pszDriver,
@@ -35,13 +36,16 @@ BOOL SQLInstallDriverEx(		LPCSTR	pszDriver,
 	char	szNameValue[INI_MAX_PROPERTY_NAME+INI_MAX_PROPERTY_VALUE+3];
 	char	szPropertyName[INI_MAX_PROPERTY_NAME+1];
 	char	szValue[INI_MAX_PROPERTY_VALUE+1];
-    char    szIniName[ INI_MAX_OBJECT_NAME + 1 ];
+    char    szIniName[ ODBC_FILENAME_MAX * 2 + 1 ];
 
 	BOOL	bInsertUsageCount;
 	int		nElement;
 	int		nUsageCount 			= 0;				/* SHOULD GET THIS FROM SOMEWHERE ? */
+	char	b1[ ODBC_FILENAME_MAX + 1 ], b2[ ODBC_FILENAME_MAX + 1 ];
 
 
+    inst_logClear();
+
 	/* SANITY CHECKS */
 	if ( pszDriver == NULL || pszPathOut == NULL )
 	{
@@ -59,17 +63,17 @@ BOOL SQLInstallDriverEx(		LPCSTR	pszDriver,
     if ( pszPathIn )
     {
 #ifdef VMS
-        sprintf( szIniName, "%sODBCINST.INI", pszPathIn );
+        snprintf( szIniName, sizeof(szIniName), "%s:%s", pszPathIn, odbcinst_system_file_name( b2 ) );
 #else
-        sprintf( szIniName, "%s/odbcinst.ini", pszPathIn );
+        snprintf( szIniName, sizeof(szIniName), "%s/%s", pszPathIn, odbcinst_system_file_name( b2 ) );
 #endif
     }
     else
     {
 #ifdef VMS
-        sprintf( szIniName, "%sODBCINST.INI", odbcinst_system_file_path() );
+        sprintf( szIniName, "%s:%s", odbcinst_system_file_path( b1 ), odbcinst_system_file_name( b2 ) );
 #else
-        sprintf( szIniName, "%s/odbcinst.ini", odbcinst_system_file_path() );
+        sprintf( szIniName, "%s/%s", odbcinst_system_file_path( b1 ), odbcinst_system_file_name( b2 ) );
 #endif
     }
 
@@ -162,13 +166,13 @@ BOOL SQLInstallDriverEx(		LPCSTR	pszDriver,
     {
         if ( pszPathOut )
         {
-            if ( strlen( odbcinst_system_file_path()) < nPathOutMax )
+            if ( strlen( odbcinst_system_file_path( b1 )) < nPathOutMax )
             {
-                strcpy( pszPathOut, odbcinst_system_file_path());
+                strcpy( pszPathOut, odbcinst_system_file_path( b1 ));
             }
             else
             {
-                strncpy( pszPathOut, odbcinst_system_file_path(), nPathOutMax );
+                strncpy( pszPathOut, odbcinst_system_file_path( b1 ), nPathOutMax );
                 pszPathOut[ nPathOutMax - 1 ] = '\0';
             }
         }
@@ -193,7 +197,7 @@ BOOL SQLInstallDriverEx(		LPCSTR	pszDriver,
     {
         if (  pszPathIn == NULL )
         {
-            *pnPathOut = strlen( odbcinst_system_file_path());
+            *pnPathOut = strlen( odbcinst_system_file_path( b1 ));
         }
         else
         {
@@ -223,6 +227,8 @@ BOOL INSTAPI SQLInstallDriverExW(LPCWSTR lpszDriver,
 	WORD len;
 	BOOL ret;
 
+    inst_logClear();
+
 	drv = lpszDriver ? _multi_string_alloc_and_copy( lpszDriver ) : (char*)NULL;
 	pth = lpszPathIn ? _single_string_alloc_and_copy( lpszPathIn ) : (char*)NULL;
 
